/**************************************************************************/
/*                                                                        */
/*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       */
/*  Kim Nguyen, Laurent Sartre                                            */
/*                                                                        */
/*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      */
/*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 */
/*                                                                        */
/*  https://www.informatique-mpi.fr/                                      */
/*                                                                        */
/**************************************************************************/

#include "prelude.h"
#include <math.h>
#include "arith.h"

int arith_gcd(int u, int v) {
  if (u < 0) u = -u;
  if (v < 0) v = -v;
  while (v != 0) {
    int tmp = v;
    v = u % v;
    u = tmp;
  }
  return u;
}

int arith_extended_gcd(int u, int v, int *x, int *y) {
  int a = 1, b = 0, c = 0, d = 1;
  if (u < 0) { u = -u; a = -1; }
  if (v < 0) { v = -v; d = -1; }
  // invariant : u = a*u0 + bv0 et v = c*u0 + d*v0
  while (v != 0) {
    int q = u / v, r = u % v;
    u = v; v = r;
    int olda = a, oldb = b;
    a = c; b = d;
    c = olda - q * c; d = oldb - q * d;
  }
  *x = a; *y = b;
  return u;
}

int arith_power_naive(int x, int n) {
  int y = 1;
  for (int i = 0; i < n; i++) {
    y *= x;
  }
  return y;
}

int arith_power(int x, int n) {
  int y = 1;
  while (n > 0) {
    if (n % 2 == 1) y *= x;
    x = x * x;
    n = n / 2;
  }
  return y;
}

uint64_t arith_power_mod(uint64_t x, uint64_t n, uint64_t m) {
  uint64_t y = 1;
  while (n > 0) {
    if (n % 2 == 1) y = (y * x) % m;
    x = (x * x) % m;
    n = n / 2;
  }
  return y;
}

uint64_t arith_inv_mod(uint64_t x, uint64_t m) {
  int y, z;
  int g = arith_extended_gcd(x, m, &y, &z);
  assert (g == 1); // et 1 = y*x + z*m
  if (y < 0) y += m;
  return y;
}

int arith_int_sqrt(int n) {
  int r = 0, i = 0;
  while (r <= n) {
    r += 2 * i + 1;
    i++;
  }
  return i - 1;
}

bool *arith_sieve(int max) {
  bool *prime = calloc(max + 1, sizeof(bool));
  for (int i = 2; i <= max; i++)
    prime[i] = true;
  for (int n = 2; n * n <= max; n++)
    if (prime[n])
      for (int m = n * n; m <= max; m += n)
        prime[m] = false;
  return prime;
}

int *arith_first_n_primes(int n) {
  int m = n < 6 ? 6 : n;
  int max = (int)(m * (log(m) + log(log(m))));
  bool *prime = arith_sieve(max);
  int *res = calloc(n, sizeof(int));
  int next = 0;
  for (int i = 2; next < n; i++)
    if (prime[i])
      res[next++] = i;
  free(prime);
  return res;
}

// test de primalité probabiliste

bool arith_fermat(uint32_t n, int k) {
  if (n <= 1) return false;
  if (n == 2 || n == 3) return true;
  if (n % 2 == 0) return false;
  for (int i = 1; i <= k ; i++) {
    int a = 2 + rand() % (n-2); // a dans [2, n-1]
    if (arith_power_mod(a, n-1, n) != 1)
      return false; // n est composé
  }
  return true; // n est probablement premier
}

bool arith_miller_rabin_check(uint32_t n, int a) {
  uint64_t s = 0, d = n - 1;
  while (d % 2 == 0) { s++; d /= 2; }
  uint64_t x = arith_power_mod(a, d, n);
  if (x == 1 || x == n-1) return true;
  for (int i = 1; i <= s-1; i++) {
    x = (x * x) % n;
    if (x == n-1) return true;
  }
  return false; // n est composé
}
bool arith_miller_rabin(uint32_t n, int k) {
  if (n <= 1) return false;
  if (n == 2 || n == 3) return true;
  if (n % 2 == 0) return false;
  for (int i = 1; i <= k ; i++) {
    int a = 2 + rand() % (n-3); // a dans [2, n-2]
    if (!arith_miller_rabin_check(n, a))
      return false; // n est composé
  }
  return true; // n est probablement premier
}

// Note : bien entendu, on obtient plus simplement, et plus efficacement,
// la division euclidienne d'entiers positifs avec / et % directement.
// Il s'agit ici d'illustrer le renvoi d'une structure contenant deux entiers.
euclidean arith_division(int a, int b) {
  assert(a >= 0 && b > 0);
  int q = 0;
  while (a >= b) {
    a -= b;
    q++;
  }
  euclidean e = { q, a };
  return e;
}

